#lang scheme

(require "board.ss"
         "find.ss"
         "test.ss")

(provide play)

;; find-new : hash-of-room hash-of-dest -> dest-or-#f
;;  Eliminated hash maps eliminated to #f (not #t!)
;;  Look for a move that takes us to a not-yet-eliminated
;;  room
(define (find-new eliminated dests)
  (for/or ([pos (in-hash-keys dests)])
    (and (symbol? (car pos))
         (hash-ref eliminated (car pos) pos))))
(test (find-new #hash() #hash()) #f)
(test (find-new #hash() #hash(((1 1) . #t))) #f)
(test (find-new #hash() #hash(((1 1) . #t) ((U S) . #t))) '(U S))
(test (find-new #hash((U . #f)) #hash(((1 1) . #t) ((U S) . #t))) #f)

(define (should-roll? eliminated state me random?)
  (define (none? v) (equal? v #hash()))
  (let-values ([(roll-destss  nonroll-dests)
                (find-all-possible state me)])
    (cond
      ;; No way to move if we don't roll? Then roll.
      [(none? nonroll-dests) #t]
      ;; No way t move it we roll? Then don't roll.
      [(andmap none? roll-destss) #f]
      ;; Randomly pick, unless we're sensible
      [random? (zero? (random 2))]
      ;; More sensibly, don't roll if you can move
      ;;  directly to a suspicious room (via a secret passage):
      [(find-new eliminated nonroll-dests) #f]
      ;; Otherwise, just roll:
      [else #t])))

(define (know-answer? eliminated)
  (define (count l)
    (for/fold ([c 0]) ([v (in-list l)])
      (+ c (if (hash-ref eliminated v #t) 1 0))))
  (and (= (count all-players) 1)
       (= (count all-weapons) 1)
       (= (count all-rooms) 1)))
(test (know-answer? #hash((U . #f) (0 . #f) (candlestick . #f))) #f)
(test (know-answer? (make-immutable-hash (map (lambda (x) (cons x #f))
                                              (append (cdr all-players)
                                                      (cdr all-weapons)
                                                      (cdr all-rooms)))))
      #t)

(define (make-eliminated-all-else who what where)
  (make-immutable-hash (map (lambda (x) (cons x #f))
                            (append (remove who all-players)
                                    (remove what all-weapons)
                                    (remove where all-rooms)))))

(define (pick-absent l eliminated)
  (random-ref
   (for/list ([v (in-list l)]
              #:when (hash-ref eliminated v #t))
     v)))

(define (finish-game r w eliminated)
  (say w 'accuse)
  (receive r 'accuse)
  (say w (list (pick-absent all-players eliminated)
               (pick-absent all-weapons eliminated)
               (pick-absent all-rooms eliminated))))

;; receive : input-port symbol -> any
;;  Receives a particular expected message, and returns
;;  the data produced by `read'
(define (receive r expected-msg)
  (let ([got-msg (read r)])
    (if (eq? got-msg expected-msg)
        (read r)
        (error 'receive "expected ~s; got ~s" expected-msg got-msg))))

(define (move eliminated state me roll random?)
  (let ([dests (find-dests state me roll)])
    (if (zero? (hash-count dests))
        ;; We can't move
        (list-ref state me)
        (or
         ;; Look for a move that takes us to a suspicious room:
         (and (not random?) (find-new eliminated dests))
         ;; Otherwise, pick randomly:
         (random-ref
          (for/list ([pos (in-hash-keys dests)]) pos))))))

;; say : output-port any -> void
;;  Writes the given reply
(define (say w what)
  (fprintf w "~a\n" what)
  (flush-output w))

(define (play host port name random?)
  (let-values ([(r w) (tcp-connect host port)])
    ;; Register our name:
    (say w name)
    ;; Get our position:
    (let ([me (receive r 'hello)])
      ;; Loop, keeping track of players, weapons, and rooms
      ;; that we've eliminated:
      (let loop ([eliminated #hash()])
        ;; Wait until our turn:
        (let ([state (receive r 'turn)])
          (cond
            [(know-answer? eliminated)
             ;; Win!
             (finish-game r w eliminated)]
            [else
             ;; Normal turn
             (say w (if (should-roll? eliminated state me random?)                       
                        'roll
                        'no-roll))
             (let* ([roll (receive r 'rolled)]
                    [dest (move eliminated state me roll random?)])
               (say w dest)
               (cond
                 [(and (symbol? (car dest))
                       (not (equal? dest (list-ref state me))))
                  ;; We entered a room, so make a suggestion:
                  (let ([here (receive r 'suggest)])
                    ;; Assert: (eq? here (car dest))
                    (let ([who (pick-absent all-players eliminated)]
                          [what (pick-absent all-weapons eliminated)]
                          [where here])
                      (say w (list who what where))
                      (let* ([counter (receive r 'counter)])
                        (if (eq? counter '-)
                            ;; We guessed right! Win on next turn:
                            (loop (make-eliminated-all-else who what where))
                            ;; Something was eliminated:
                            (loop (hash-set eliminated counter #f))))))]
                 [else
                  ;; not in a room:
                  (loop eliminated)]))]))))))
